"""
    Diagnostic base classes
    =======================

    Abstract base classes defining the diagnostics of the model and used to analyze its outputs.

    Description of the classes
    --------------------------

    * :class:`Diagnostic`: General base class.
    * :class:`FieldDiagnostic`: General base class for diagnostics returning model's fields.

    Warnings
    --------

    These are `abstract base class`_, they must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

"""

from abc import ABC, abstractmethod
import warnings
import base64

import matplotlib.pyplot as plt
import matplotlib.animation as animation
from IPython.display import HTML, display
from ipywidgets import interactive

from sympy import symbols

from qgs.diagnostics.misc import *

# TODO: - need to introduce an oro_basis specific to the orography !!
#       - no orography plot when oro_basis is none


_n = symbols('n', real=True, nonnegative=True)  # only this works with sympy substitution, don't ask me why...


class Diagnostic(ABC):
    """General base class to create diagnostics.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    """

    def __init__(self, model_params, dimensional):

        self._model_params = None
        self._ocean = None
        self._ground = None
        self._orography = None
        self._heat_exchange = None
        self._newton = None
        self._data = None
        self._diagnostic_data = None
        self._diagnostic_data_dimensional = None
        self._time = None
        self._subs = list()

        self.dimensional = dimensional

        self._plot_title = ""
        self._plot_units = ""

        self.set_params(model_params)

    def __call__(self, time, data):
        self.set_data(time, data)
        return self.diagnostic

    @property
    def diagnostic(self):
        """~numpy.ndarray: The output diagnostic."""
        diag = self._check_diagnostic(self.dimensional)
        if diag is False:
            return self._get_diagnostic(self.dimensional)
        else:
            return diag

    def _check_diagnostic(self, dimensional):

        if self._data is None:
            return None
        elif self._diagnostic_data is not None and self._diagnostic_data_dimensional == dimensional:
            return self._diagnostic_data
        elif self._data.shape[0] != self._model_params.ndim:
            warnings.warn('Problem with the provided data. Expected array with 0-axis shape of '+str(self._model_params.ndim)+' .'
                          + 'Got ' + str(self._data.shape[0]) + ' ! Returning None.')
            return None
        else:
            return False

    def set_params(self, model_params, kwargs=None):
        """Set or replace the current model's parameter to which the diagnostic is attached to.

        Parameters
        ----------

        model_params: QgParams
            An instance of the model parameters.
        kwargs: dict
            Arguments to eventually reconfigure the diagnostic instance. Have the same keywords as the diagnostic instantiation method.
            If `None`, does not reconfigure the instance.
            Default to `None`.
        """

        self._set_params(model_params)
        self._subs = [(_n, model_params.scale_params.n)]
        if kwargs is not None:
            self._configure(**kwargs)

    def _set_params(self, model_params):

        self._model_params = model_params

        # determine some parameters about the model

        self._ocean = model_params.oceanic_basis is not None
        self._ground = model_params.ground_basis is not None
        if model_params.ground_params is not None:
            self._orography = model_params.ground_params.hk is not None
        self._heat_exchange = model_params.atemperature_params.C is not None
        self._newton = model_params.atemperature_params.thetas is not None

    def set_data(self, time, data):
        """Provide the model data to the diagnostic.

        Parameters
        ----------
        time: ~numpy.ndarray
            The time (in nondimensional timeunits) corresponding to the data.
            Its length should match the length of the last axis of the provided `data`.
        data: ~numpy.ndarray
            The model output data that the user want to convert using the diagnostic.
            Should be a 2D array of shape (:attr:`~.params.QgParams.ndim`, number_of_timesteps).
        """

        self._data = data
        self._time = time
        self._diagnostic_data = None

    @abstractmethod
    def _get_diagnostic(self, dimensional):
        pass

    @abstractmethod
    def _configure(self, **kwargs):
        pass


class FieldPointDiagnostic(Diagnostic):
    """General base class to give field values over time at a given point of the domain.

    Warnings
    --------

    Not yet implemented.

    """

    # TODO: WIP, still to be implemented

    def __init__(self, model_params, dimensional):

        Diagnostic.__init__(self, model_params, dimensional)

        self._x = None
        self._y = None

    def set_point_coordinates(self, x, y):

        self._x = x
        self._y = y

    def _get_point_coordinates(self):
        return self._x, self._y

    @property
    def point_coordinates(self):
        return self._get_point_coordinates()


class FieldDiagnostic(Diagnostic):
    """General base class for field diagnostic on the model's domain.
    Should provide a spatial gridded representation of the fields.


    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, dimensional):

        Diagnostic.__init__(self, model_params, dimensional)

        self._X = None
        self._Y = None

        self._grid_basis = None
        self._oro_basis = None

        self._color_bar_format = True

        self._default_plot_kwargs = {'cmap': plt.get_cmap('jet'), 'interpolation': 'spline36'}

    def __len__(self):
        if self.diagnostic is not None:
            return self.diagnostic.shape[0]
        else:
            return None

    @abstractmethod
    def _compute_grid(self, delta_x=None, delta_y=None):
        pass

    @property
    def grid_shape(self):
        """tuple(int): Return the shape of the grid of points covering the model's domain."""
        if self._Y is not None:
            return self._Y.shape

    def plot(self, time_index, style="image", ax=None, figsize=(16, 9),
             contour_labels=True, color_bar=True, show_time=True, plot_kwargs=None, oro_kwargs=None):
        """Plot the field of the provided data at the given time index.

        Parameters
        ----------
        time_index: int
            The time index of the data.
        style: str, optional
            The style of the plot. Can be:

            * `image`: show the fields as images with a given colormap specified in the `plot_kwargs` argument.
            * `contour`: show the fields as contour superimposed on the image of the orographic height (if it exists, see the `oro_kwargs` below).
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        contour_labels: bool, optional
            If `style` is set to `contour`, specify if the contours must be labelled with their value or not.
            Default to `True`.
        color_bar: bool, optional
            Specify if a color bar must be drawn beside the plot or not.
            Default to `True`.
        show_time: bool, optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        plot_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.
        oro_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method plotting the image of the orography if `style` is set to `contour`.

        Returns
        -------
        ~matplotlib.axes.Axes
            An axes where the data were plotted.
        """

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing.')
            return None

        if time_index > self.__len__() - 1:
            warnings.warn('Time index ' + str(time_index) + ' provided is greater than the largest one possible: ' + str(self.__len__()) + ' . Showing nothing.')
            return None

        if plot_kwargs is None:
            plot_kwargs = self._default_plot_kwargs
        else:
            tmp_dict = dict(self._default_plot_kwargs)
            tmp_dict.update(plot_kwargs)
            plot_kwargs = tmp_dict

        if oro_kwargs is None:
            oro_kwargs = dict()
            oro_kwargs['cmap'] = plt.get_cmap('cividis')
            oro_kwargs['alpha'] = 0.6
            oro_kwargs['interpolation'] = 'spline36'

        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.gca()
        else:
            fig = ax.figure

        if style == "image":
            im = ax.imshow(self.diagnostic[time_index], origin='lower',
                           extent=[0, 2*np.pi/self._model_params.scale_params.n, 0, np.pi], **plot_kwargs)
            if color_bar:
                if self._color_bar_format:
                    cl = fig.colorbar(im, ax=ax, format=strf)
                else:
                    cl = fig.colorbar(im, ax=ax)

        elif style == "contour":
            if self._orography and oro_kwargs is not False and self._oro_basis is not None:
                hk = np.array(self._model_params.ground_params.hk, dtype=float)
                oro = hk @ np.swapaxes(self._oro_basis, 0, 1) * self._model_params.scale_params.Ha
                im = ax.imshow(oro, origin='lower',
                               extent=[0, 2 * np.pi / self._model_params.scale_params.n, 0, np.pi], **oro_kwargs)
                if color_bar:
                    cl = fig.colorbar(im, ax=ax)
                    cl.set_label('Orographic height (in m)', rotation=270, labelpad=20.)
            if 'cmap' in plot_kwargs:
                del plot_kwargs['cmap']
            if 'interpolation' in plot_kwargs:
                del plot_kwargs['interpolation']
            cont = ax.contour(self._X, self._Y, self.diagnostic[time_index], **plot_kwargs)
            if contour_labels:
                ax.clabel(cont, fontsize=10, fmt='%1.f')
        else:
            warnings.warn('Provided style parameter ' + style + ' not supported ! Nothing to plot.')

        title = self._plot_title
        if self.dimensional:
            title += self._plot_units
            if show_time:
                title += " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[time_index]) + " " + self._model_params.time_unit
        else:
            title += r" (in nondim units)"
            if show_time:
                title += " at " + str(self._time[time_index]) + " timeunits"

        ax.set_title(title, pad=20.)

        ax.set_xlabel(x_axis_label)
        ax.set_ylabel(y_axis_label)

        if oro_kwargs is False:
            color_bar = False

        if color_bar:
            return ax, cl
        else:
            return ax

    def plot_grid_point(self, i, j, ax=None, figsize=(16, 9), plot_kwargs=None):
        """Plot the time serie of the field at a given grid point.

        Attributes
        ----------
        i: int
            Index corresponding to the x-coordinate of the grid point.
        j: int
            Index corresponding to the y-coordinate of the grid point.
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        plot_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.plot` method.

        Returns
        -------
        ~matplotlib.axes.Axes
            An axes where the data were plotted.
        """

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        if plot_kwargs is None:
            plot_kwargs = dict()

        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.gca()
        else:
            fig = ax.figure

        ax.plot(self._time * (1 - int(self.dimensional) + int(self.dimensional) * self._model_params.dimensional_time),
                self.diagnostic[:, j, i], **plot_kwargs)

        title = self._plot_title
        if self.dimensional:
            title += self._plot_units
        else:
            title += r" (in nondim units)"
        title += " at gridpoint " + "({:.2f}".format(self._X[j, i]) + ", {:.2f}) ".format(self._Y[j, i])

        ax.set_title(title, pad=20.)

        if self.dimensional:
            ax.set_xlabel(time_axis_label + ' [' + self._model_params.time_unit + ']')
        else:
            ax.set_xlabel(time_axis_label + ' [timeunits]')

        return ax

    def movie(self, output='html', filename='', style="image", ax=None, figsize=(16, 9),
              contour_labels=True, color_bar=True, show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None):
        """ Create and return a movie of the output of the `plot` method animated over time.

        Parameters
        ----------
        output: str, optional
            Define the kind of movie being created. Can be:

            * `jshtml`: Generate an interactive HTML representation of the animation.
            * `html5`: Generate the movie as HTML5 code.
            * `html`: Output the movie as a HTML video tag.
            * `ihtml`: Output the interactive movie as a HTML video tag.
            * `save`: Save the movie in MP4 format (H264 codec).

            Default to `html`.
        filename: str, optional
            Filename (and path) where to save the movie. Needed if `output` is set to `save`.
        style: str, optional
            The style of the plot. Can be:

            * `image`: show the fields as images with a given colormap specified in the `plot_kwargs` argument.
            * `contour`: show the fields as contour superimposed on the image of the orographic height (see the `oro_kwargs` below).
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        contour_labels: bool, optional
            If `style` is set to `contour`, specify if the contours must be labelled with their value or not.
            Default to `True`.
        color_bar: bool, optional
            Specify if a color bar must be drawn beside the plot or not.
            Default to `True`.
        show_time: bool, optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        plot_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.
        oro_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method plotting the image of the orography if `style` is set to `contour`.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or HTML code or HTML tag
            The animation object or the HTML code or tag.
        """

        anim = self._make_anim(style, ax, figsize, contour_labels, color_bar, show_time, plot_kwargs, oro_kwargs, anim_kwargs, False)

        if 'html' in output:

            if output == "jshtml" or output == 'ihtml':
                jshtml = anim.to_jshtml()
                if output == "jshtml":
                    return jshtml
                else:
                    return HTML(jshtml)
            else:
                html5 = anim.to_html5_video()
                if output == 'html5':
                    return html5
                else:
                    return HTML(html5)

        elif output == 'save':

            if not filename:
                warnings.warn('No filename provided to the method animate. Video not saved !\n Please provide a filename.')

            html5 = anim.to_html5_video()
            start_index = html5.index('base64,')
            start_index += len('base64,')
            end_index = html5.index('">', start_index)
            video = html5[start_index: end_index]
            with open(filename, 'wb') as f:
                f.write(base64.b64decode(video))
            return html5

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def animate(self, output='animate',  style="image", ax=None, figsize=(16, 9),
                contour_labels=True, color_bar=True, show_time=True, stride=1, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None, show=True):
        """Return the output of the `plot` method animated over time.

        Parameters
        ----------
        output: str, optional
            Define the kind of animation being created. Can be:

            * `animate`: Create and show a :class:`ipywidgets.widgets.interaction.interactive` widget. Works only in Jupyter notebooks.
            * `show`: Create and show an animation with the :mod:`matplotlib.animation` module. Works only in IPython or Python.

        style: str, optional
            The style of the plot. Can be:

            * `image`: show the fields as images with a given colormap specified in the `plot_kwargs` argument.
            * `contour`: show the fields as contour superimposed on the image of the orographic height (see the `oro_kwargs` below).

        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        contour_labels: bool, optional
            If `style` is set to `contour`, specify if the contours must be labelled with their value or not.
            Default to `True`.
        color_bar: bool, optional
            Specify if a color bar must be drawn beside the plot or not.
            Default to `True`.
        show_time: bool, optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        stride: int, optional
            Specify the time step of the animation. Works only with `output` set to `animate`.
        plot_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.
        oro_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method plotting the image of the orography if `style` is set to `contour`.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.
            Works only with `output` set to `show`.
        show: bool, optional
            Whether to plot or not the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or ~IPython.display.DisplayHandle or callable
            The animation object or the callable to update the widget, depending on the value of the `output` and `show` parameters.
        """

        if output == 'animate':

            if self.diagnostic is None:
                warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
                return None

            if style == 'image':
                vmin = self.diagnostic.min() * 1.03
                vmax = self.diagnostic.max() * 1.03
                if plot_kwargs is not None:
                    if 'vmin' not in plot_kwargs:
                        plot_kwargs['vmin'] = vmin
                    if 'vmax' not in plot_kwargs:
                        plot_kwargs['vmax'] = vmax
                else:
                    plot_kwargs = dict()
                    plot_kwargs['vmin'] = vmin
                    plot_kwargs['vmax'] = vmax

            def update(time_index):
                self.plot(time_index, style, ax, figsize, contour_labels, color_bar, show_time, plot_kwargs, oro_kwargs)
                if show:
                    plt.show()

            if show:
                plot = interactive(update, time_index=(0, len(self)-1, stride))
                anim = display(plot)
            else:
                return update

        elif output == 'show':

            anim = self._make_anim(style, ax, figsize, contour_labels, color_bar, show_time, plot_kwargs, oro_kwargs, anim_kwargs, True)
            if show:
                plt.show()

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def _init_anim(self, style="image", ax=None, figsize=(16, 9), contour_labels=True, color_bar=True,
                   show_time_in_title=False, show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None):

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        if style == 'image':
            vmin = self.diagnostic.min() * 1.03
            vmax = self.diagnostic.max() * 1.03
            if plot_kwargs is not None:
                if 'vmin' not in plot_kwargs:
                    plot_kwargs['vmin'] = vmin
                if 'vmax' not in plot_kwargs:
                    plot_kwargs['vmax'] = vmax
            else:
                plot_kwargs = dict()
                plot_kwargs['vmin'] = vmin
                plot_kwargs['vmax'] = vmax

        pack = self.plot(0, style, ax, figsize, contour_labels, color_bar, show_time_in_title, plot_kwargs, oro_kwargs)
        if hasattr(pack, '__getitem__'):
            ax = pack[0]
        else:
            ax = pack
        fig = ax.figure

        if show_time:
            if self.dimensional:
                tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[0]) + " " + self._model_params.time_unit
            else:
                tt = " at " + str(self._time[0]) + " timeunits"
            ax.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

        fargs = (ax, show_time_in_title, show_time)

        kwargs = {'style': style, 'ax': ax, 'figsize': figsize, 'contour_labels': contour_labels, 'color_bar': color_bar,
                  'show_time_in_title': show_time_in_title, 'show_time': show_time, 'plot_kwargs': plot_kwargs, 'oro_kwargs': oro_kwargs,
                  'anim_kwargs': anim_kwargs}

        return fig, ax, fargs, kwargs

    def _make_update(self, style="image", ax=None, figsize=(16, 9), contour_labels=True, color_bar=True,
                     show_time_in_title=False, show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None):
        def update(i, axe, show_t_in_title, show_t):
            axe.clear()
            axe = self.plot(i, style, axe, figsize, contour_labels, False, show_t_in_title, plot_kwargs, oro_kwargs)
            if show_t:
                if self.dimensional:
                    tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[i]) + " " + self._model_params.time_unit
                else:
                    tt = " at " + str(self._time[i]) + " timeunits"
                axe.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=axe.transAxes)
            return [axe]

        return update

    def _make_anim(self, style="image", ax=None, figsize=(16, 9), contour_labels=True, color_bar=True,
                   show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None, blit=True):

        if show_time:
            if blit:
                show_time_in_title = False
            else:
                show_time_in_title = True
                show_time = False
        else:
            show_time_in_title = False

        fig, ax, fargs, kwargs = self._init_anim(style, ax, figsize, contour_labels, color_bar, show_time_in_title, show_time, plot_kwargs, oro_kwargs, anim_kwargs)

        update = self._make_update(**kwargs)

        if anim_kwargs is not None:

            if 'blit' in anim_kwargs:
                del anim_kwargs['blit']

            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit, **anim_kwargs)

        else:
            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit)

        return anim


class ProfileDiagnostic(Diagnostic):
    """General base class for profile diagnostic on the model's domain.
    Should provide a 1D representation of the fields averages or section.


    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, dimensional):

        Diagnostic.__init__(self, model_params, dimensional)

        self._points_coordinates = None
        self._plot_label = None
        self._axis_label = None

    # self._default_plot_kwargs = {'cmap': plt.get_cmap('jet'), 'interpolation': 'spline36'}

    def __len__(self):
        if self.diagnostic is not None:
            return self.diagnostic.shape[0]
        else:
            return None

    def plot(self, time_index=0, ax=None, figsize=(16, 9), show_time=True, plot_kwargs=None, **kwargs):
        """Plot the multiple profile diagnostic provided.

        Parameters
        ----------
        time_index: int
            The time index of the data. Not used in this subclass.
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        show_time: bool, optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        plot_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.

        Returns
        -------
        ~matplotlib.axes.Axes
            An axes where the data were plotted.
        """

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        if time_index > self.__len__() - 1:
            warnings.warn('Time index ' + str(time_index) + ' provided is greater than the largest one possible: ' + str(self.__len__()) + ' . Showing nothing.')
            return None

        if plot_kwargs is None:
            plot_kwargs = dict()

        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(1, 1, 1)
        else:
            fig = ax.figure

        if 'label' not in plot_kwargs:
            lab = self._plot_label
            if self.dimensional:
                lab += r" [" + self._plot_units + r"]"
            plot_kwargs = {'label': lab}

        ax.plot(self._points_coordinates, self.diagnostic[time_index], **plot_kwargs)
        ax.legend()

        ax.set_xlabel(self._axis_label)

        title = self._plot_title
        if self.dimensional:
            if show_time:
                title += " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[time_index]) + " " + self._model_params.time_unit
        else:
            title += r" (in nondim units)"
            if show_time:
                title += " at " + str(self._time[time_index]) + " timeunits"

        ax.set_title(title, pad=20.)

        return ax

    def movie(self, output='html', filename='', ax=None, figsize=(16, 9), plot_kwargs=None, anim_kwargs=None):
        """ Create and return a movie of the output of the `plot` method animated over time.
        Show a red dot moving and depicting the current value of the model's selected variables.

        Parameters
        ----------
        output: str, optional
            Define the kind of movie being created. Can be:

            * `jshtml`: Generate an interactive HTML representation of the animation.
            * `html5`: Generate the movie as HTML5 code.
            * `html`: Output the movie as a HTML video tag.
            * `ihtml`: Output the interactive movie as a HTML video tag.
            * `save`: Save the movie in MP4 format (H264 codec).

            Default to `html`.
        filename: str, optional
            Filename (and path) where to save the movie. Needed if `output` is set to `save`.
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot.
            If not provided, create a new one.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        plot_kwargs: dict, optional
            Arguments to pass to the background plot.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or HTML code or HTML tag
            The animation object or the HTML code or tag.
        """

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        anim = self._make_anim(ax, figsize, True, plot_kwargs, anim_kwargs, False)

        if 'html' in output:

            if output == "jshtml" or output == 'ihtml':
                jshtml = anim.to_jshtml()
                if output == "jshtml":
                    return jshtml
                else:
                    return HTML(jshtml)
            else:
                html5 = anim.to_html5_video()
                if output == 'html5':
                    return html5
                else:
                    return HTML(html5)

        elif output == 'save':

            if not filename:
                warnings.warn('No filename provided to the method animate. Video not saved !\n Please provide a filename.')

            html = anim.to_html5_video()
            start_index = html.index('base64,')
            start_index += len('base64,')
            end_index = html.index('">', start_index)
            video = html[start_index: end_index]
            with open(filename, 'wb') as f:
                f.write(base64.b64decode(video))

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def animate(self, output='animate', ax=None, figsize=(16, 9), show_time=True, stride=1, plot_kwargs=None, anim_kwargs=None, show=True):
        """Return the output of the `plot` method animated over time.

        Parameters
        ----------
        output: str, optional
            Define the kind of animation being created. Can be:

            * `animate`: Create and show a :class:`ipywidgets.widgets.interaction.interactive` widget. Works only in Jupyter notebooks.
            * `show`: Create and show an animation with the :mod:`matplotlib.animation` module. Works only in IPython or Python.

        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        show_time: bool, optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        stride: int, optional
            Specify the time step of the animation. Works only with `output` set to `animate`.
        plot_kwargs: dict, optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.
            Works only with `output` set to `show`.
        show: bool, optional
            Whether to plot or not the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or ~IPython.display.DisplayHandle or callable
            The animation object or the callable to update the widget, depending on the value of the `output` and `show` parameters.
        """

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        if output == 'animate':

            heat_max = np.max(self.diagnostic)
            heat_min = np.min(self.diagnostic)

            def update(time_index):
                self.plot(time_index, ax, figsize, show_time, plot_kwargs)
                ax.set_ylim([heat_min, heat_max])
                ax.legend(loc=1)
                if show:
                    plt.show()

            if show:
                plot = interactive(update, time_index=(0, len(self)-1, stride))
                anim = display(plot)
            else:
                return update

        elif output == 'show':

            anim = self._make_anim(ax, figsize, show_time, plot_kwargs, anim_kwargs, True)
            if show:
                plt.show()

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def _init_anim(self, ax=None, figsize=(16, 9), show_time_in_title=False, show_time=True, plot_kwargs=None, anim_kwargs=None):

        if self.diagnostic is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        heat_max = np.max(self.diagnostic)
        heat_min = np.min(self.diagnostic)

        ax = self.plot(0, ax, figsize, show_time_in_title, plot_kwargs)
        fig = ax.figure
        ax.set_ylim([heat_min, heat_max])
        ax.legend(loc=1)

        if show_time:
            if self.dimensional:
                tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[0]) + " " + self._model_params.time_unit
            else:
                tt = " at " + str(self._time[0]) + " timeunits"
            ax.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

        fargs = (ax, show_time_in_title, show_time)

        kwargs = {'ax': ax, 'figsize': figsize, 'show_time_in_title': show_time_in_title, 'show_time': show_time, 'plot_kwargs': plot_kwargs,
                  'anim_kwargs': anim_kwargs}

        return fig, ax, fargs, kwargs

    def _make_update(self, ax=None, figsize=(16, 9), show_time_in_title=False, show_time=True, plot_kwargs=None, anim_kwargs=None):

        heat_max = np.max(self.diagnostic)
        heat_min = np.min(self.diagnostic)

        def update(i, axe, show_t_in_title, show_t):
            axe.clear()
            axe = self.plot(i, axe, figsize, show_t_in_title, plot_kwargs)
            axe.set_ylim([heat_min, heat_max])
            axe.legend(loc=1)
            if show_t:
                if self.dimensional:
                    tt = " at " + "{:.2f}".format(self._model_params.dimensional_time * self._time[i]) + " " + self._model_params.time_unit
                else:
                    tt = " at " + str(self._time[i]) + " timeunits"
                axe.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=axe.transAxes)
            return [axe]

        return update

    def _make_anim(self, ax=None, figsize=(16, 9), show_time=True, plot_kwargs=None, anim_kwargs=None, blit=True):

        if show_time:
            if blit:
                show_time_in_title = False
            else:
                show_time_in_title = True
                show_time = False
        else:
            show_time_in_title = False

        fig, ax, fargs, kwargs = self._init_anim(ax, figsize, show_time_in_title, show_time, plot_kwargs, anim_kwargs)

        update = self._make_update(**kwargs)

        if anim_kwargs is not None:

            if 'blit' in anim_kwargs:
                del anim_kwargs['blit']

            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit, **anim_kwargs)

        else:
            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit)

        return anim



"""
    Diagnostic temperature classes
    =================================

    Classes defining temperature fields diagnostics.

    Description of the classes
    --------------------------

    * :class:`AtmosphericTemperatureDiagnostic`: General base class for atmospheric temperature fields diagnostics.
    * :class:`MiddleAtmosphericTemperatureDiagnostic`: Diagnostic giving the middle atmospheric anomaly fields :math:`T_{\\rm a}`.
    * :class:`MiddleAtmosphericTemperatureAnomalyDiagnostic`: Diagnostic giving the middle atmospheric temperature  anomaly fields :math:`\\delta T_{\\rm a}`.
    * :class:`OceanicTemperatureDiagnostic`: General base class for oceanic temperature fields diagnostic.
    * :class:`OceanicLayerTemperatureDiagnostic`: Diagnostic giving the oceanic layer temperature fields :math:`T_{\\rm o}`.
    * :class:`OceanicLayerTemperatureAnomalyDiagnostic`: Diagnostic giving the oceanic layer temperature anomaly fields :math:`\\delta T_{\\rm o}`.
    * :class:`GroundTemperatureDiagnostic`: Diagnostic giving the ground layer temperature fields :math:`T_{\\rm g}`.
    * :class:`GroundTemperatureAnomalyDiagnostic`: Diagnostic giving the ground layer temperature anomaly fields :math:`\\delta T_{\\rm g}`.
    * :class:`AtmosphericTemperatureMeridionalGradientDiagnostic`: General base class for meridional gradient of atmospheric temperature fields diagnostics.
    * :class:`MiddleAtmosphericTemperatureMeridionalGradientDiagnostic`: Diagnostic giving the meridional gradient of the middle atmospheric temperature fields :math:`\\partial_y T_{\\rm a}`.

"""
import warnings

import numpy as np
import matplotlib.pyplot as plt
from qgs.diagnostics.util import create_grid_basis

from qgs.diagnostics.base import FieldDiagnostic
from qgs.diagnostics.differential import DifferentialFieldDiagnostic


class AtmosphericTemperatureDiagnostic(FieldDiagnostic):
    """General base class for atmospheric temperature fields diagnostic.
    Provide a spatial gridded representation of the fields.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        FieldDiagnostic.__init__(self, model_params, dimensional)
        self._configure(delta_x=delta_x, delta_y=delta_y)
        self._default_plot_kwargs['cmap'] = plt.get_cmap('coolwarm')

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericTemperatureDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [ams[i][0] for i in range(len(ams))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericTemperatureDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [ams[i][1] for i in range(len(ams))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, model_params=None, delta_x=None, delta_y=None):

        self._compute_grid(delta_x, delta_y)

        self._grid_basis = create_grid_basis(self._model_params.atmospheric_basis, self._X, self._Y, self._subs)
        if self._orography:
            if self._model_params.ground_params.orographic_basis == "atmospheric":
                self._oro_basis = self._grid_basis
            else:
                self._oro_basis = create_grid_basis(self._model_params.ground_basis, self._X, self._Y, self._subs)
        else:
            self._oro_basis = None


class MiddleAtmosphericTemperatureAnomalyDiagnostic(AtmosphericTemperatureDiagnostic):
    """Diagnostic giving the middle atmospheric temperature anomaly fields :math:`\\delta T_{\\rm a}` at 500hPa.
    It is identified with the baroclinic streamfunction :math:`\\theta_{\\rm a}` of the system.
    See also :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericTemperatureDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        vr = self._model_params.variables_range
        self._plot_title = r'Atmospheric 500hPa temperature anomaly'
        self._plot_units = r" (in " + self._model_params.get_variable_units(vr[0]) + r")"

        self._color_bar_format = False

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        theta = np.swapaxes(self._data[vr[0]+offset:vr[1], ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = theta * self._model_params.temperature_scaling * 2
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = theta
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class MiddleAtmosphericTemperatureDiagnostic(AtmosphericTemperatureDiagnostic):
    """Diagnostic giving the middle atmospheric temperature fields :math:`T_{\\rm a} = T_{{\\rm a}, 0} + \\delta T_{\\rm a}` at 500hPa,
    where :math:`T_{{\\rm a}, 0}` is the reference temperature :attr:`~.params.params.AtmosphericTemperatureParams.T0` or the 0-th order dynamic temperature.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Notes
    -----

    Only works if the heat exchange scheme is activated, i.e. does not work with the Newton cooling scheme.

    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericTemperatureDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        vr = self._model_params.variables_range
        self._plot_title = r'Atmospheric 500hPa temperature'
        self._plot_units = r" (in " + self._model_params.get_variable_units(vr[0]) + r")"

        self._color_bar_format = False

        if not self._heat_exchange:
            warnings.warn('Heat exchange scheme is not activated, this diagnostic will not work properly !')

    def _get_diagnostic(self, dimensional):

        vr = self._model_params.variables_range
        T = np.swapaxes(self._data[vr[0]:vr[1], ...].T @ np.swapaxes(self._grid_basis, 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = T * self._model_params.temperature_scaling * 2
            if not self._model_params.dynamic_T:
                self._diagnostic_data += self._model_params.atemperature_params.T0
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = T
            if not self._model_params.dynamic_T:
                self._diagnostic_data += self._model_params.atemperature_params.T0 / (self._model_params.temperature_scaling * 2)
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class OceanicTemperatureDiagnostic(FieldDiagnostic):
    """General base class for atmospheric temperature fields diagnostic.
    Provide a spatial gridded representation of the fields.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        FieldDiagnostic.__init__(self, model_params, dimensional)
        self._configure(delta_x=delta_x, delta_y=delta_y)
        self._default_plot_kwargs['cmap'] = plt.get_cmap('coolwarm')

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            oms = self._model_params.oblocks
            if oms is None:
                warnings.warn("OceanicTemperatureDiagnostic: Unable to configure the grid automatically. Oceanic wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [oms[i][0] for i in range(len(oms))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            oms = self._model_params.oblocks
            if oms is None:
                warnings.warn("OceanicTemperatureDiagnostic: Unable to configure the grid automatically. Oceanic wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [oms[i][1] for i in range(len(oms))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, model_params=None, delta_x=None, delta_y=None):

        if not self._ocean:
            warnings.warn("OceanicTemperatureDiagnostic: No ocean configuration found in the provided parameters. This model version does not have an ocean. " +
                          "Please check your configuration.")
            return 1

        self._compute_grid(delta_x, delta_y)
        basis = self._model_params.oceanic_basis

        self._grid_basis = create_grid_basis(basis, self._X, self._Y, self._subs)


class OceanicLayerTemperatureAnomalyDiagnostic(OceanicTemperatureDiagnostic):
    """Diagnostic giving the oceanic layer temperature anomaly fields :math:`\\delta T_{\\rm o}`.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        OceanicTemperatureDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        vr = self._model_params.variables_range
        self._plot_title = r'Oceanic temperature anomaly'
        self._plot_units = r" (in " + self._model_params.get_variable_units(vr[2]) + r")"

        self._color_bar_format = False

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        delta_T = np.swapaxes(self._data[vr[2]+offset:, ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = delta_T * self._model_params.temperature_scaling
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = delta_T
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class OceanicLayerTemperatureDiagnostic(OceanicTemperatureDiagnostic):
    """Diagnostic giving the oceanic layer temperature fields :math:`T_{\\rm o} = T_{{\\rm o}, 0} + \\delta T_{\\rm o}`,
    where :math:`T_{{\\rm o}, 0}` is the reference temperature :attr:`~.params.params.OceanicTemperatureParams.T0` or the 0-th order dynamic temperature.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        OceanicTemperatureDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)

        vr = self._model_params.variables_range
        self._plot_title = r'Oceanic temperature'
        self._plot_units = r" (in " + self._model_params.get_variable_units(vr[2]) + r")"

        self._color_bar_format = False

    def _get_diagnostic(self, dimensional):

        vr = self._model_params.variables_range
        T = np.swapaxes(self._data[vr[2]:, ...].T @ np.swapaxes(self._grid_basis, 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = T * self._model_params.temperature_scaling
            if not self._model_params.dynamic_T:
                self._diagnostic_data += self._model_params.gotemperature_params.T0
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = T
            if not self._model_params.dynamic_T:
                self._diagnostic_data += self._model_params.gotemperature_params.T0 / self._model_params.temperature_scaling
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class GroundTemperatureAnomalyDiagnostic(FieldDiagnostic):
    """Diagnostic giving the ground temperature anomaly fields :math:`\\delta T_{\\rm g}`.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        FieldDiagnostic.__init__(self, model_params, dimensional)
        self._configure(delta_x=delta_x, delta_y=delta_y)
        self._default_plot_kwargs['cmap'] = plt.get_cmap('coolwarm')

        vr = self._model_params.variables_range
        self._plot_title = r'Ground temperature anomaly'
        self._plot_units = r" (in " + self._model_params.get_variable_units(vr[1]) + r")"

        self._color_bar_format = False

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            gms = self._model_params.gblocks
            if gms is None:
                warnings.warn("GroundTemperatureDiagnostic: Unable to configure the grid automatically. Ground wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [gms[i][0] for i in range(len(gms))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            gms = self._model_params.gblocks
            if gms is None:
                warnings.warn("GroundTemperatureDiagnostic: Unable to configure the grid automatically. Ground wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [gms[i][1] for i in range(len(gms))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, model_params=None, delta_x=None, delta_y=None):

        if not self._ground:
            warnings.warn("GroundTemperatureDiagnostic: No ground configuration found in the provided parameters. This model version does not have a ground. " +
                          "Please check your configuration.")
            return 1

        self._compute_grid(delta_x, delta_y)
        basis = self._model_params.ground_basis

        self._grid_basis = create_grid_basis(basis, self._X, self._Y, self._subs)

    def _get_diagnostic(self, dimensional):

        if self._model_params.dynamic_T:
            offset = 1
        else:
            offset = 0

        vr = self._model_params.variables_range
        delta_T = np.swapaxes(self._data[vr[1]+offset:, ...].T @ np.swapaxes(self._grid_basis[offset:], 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = delta_T * self._model_params.temperature_scaling
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = delta_T
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class GroundTemperatureDiagnostic(GroundTemperatureAnomalyDiagnostic):
    """Diagnostic giving the ground temperature fields :math:`T_{\\rm g} = T_{{\\rm g}, 0} + \\delta T_{\\rm g}`,
    where :math:`T_{{\\rm g}, 0}` is the reference temperature :attr:`~.params.params.GroundTemperatureParams.T0` or the 0-th order dynamic temperature.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        GroundTemperatureAnomalyDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)
        self._plot_title = r'Ground temperature'

    def _get_diagnostic(self, dimensional):

        vr = self._model_params.variables_range
        T = np.swapaxes(self._data[vr[1]:, ...].T @ np.swapaxes(self._grid_basis, 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = T * self._model_params.temperature_scaling
            if not self._model_params.dynamic_T:
                self._diagnostic_data += self._model_params.gotemperature_params.T0
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = T
            if not self._model_params.dynamic_T:
                self._diagnostic_data += self._model_params.gotemperature_params.T0 / self._model_params.temperature_scaling
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


class AtmosphericTemperatureMeridionalGradientDiagnostic(DifferentialFieldDiagnostic):
    """General base class for atmospheric temperature fields meridional gradient diagnostic.
    Provide a spatial gridded representation of the fields.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        DifferentialFieldDiagnostic.__init__(self, model_params, dimensional)

        self._configure(delta_x=delta_x, delta_y=delta_y)
        self._default_plot_kwargs['cmap'] = plt.get_cmap('coolwarm')

    def _compute_grid(self, delta_x=None, delta_y=None):

        if delta_x is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericTemperatureDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            xwn = [ams[i][0] for i in range(len(ams))]
            mxwn = max(xwn)
            n_point_x = 4 * mxwn + 2
        else:
            n_point_x = int(np.ceil((2 * np.pi / self._model_params.scale_params.n) / delta_x) + 1)

        if delta_y is None:
            ams = self._model_params.ablocks
            if ams is None:
                warnings.warn("AtmosphericTemperatureDiagnostic: Unable to configure the grid automatically. Atmospheric wavenumbers information not " +
                              "present in the model's parameters ! Please call the compute_grid method with the delta_x and delta_y parameters.")
                return 1
            ywn = [ams[i][1] for i in range(len(ams))]
            mywn = max(ywn)
            n_point_y = 4 * mywn + 2
        else:
            n_point_y = int(np.ceil(np.pi / delta_y) + 1)

        x = np.linspace(0., 2 * np.pi / self._model_params.scale_params.n, n_point_x)
        y = np.linspace(0., np.pi, n_point_y)
        self._X, self._Y = np.meshgrid(x, y)

    def _configure(self, delta_x=None, delta_y=None):

        basis = self._model_params.atmospheric_basis

        self._configure_differential_grid(basis, "dy", 1, delta_x, delta_y)

        if self._orography and self._X is not None and self._Y is not None:
            if self._model_params.ground_params.orographic_basis == "atmospheric":
                self._oro_basis = create_grid_basis(basis, self._X, self._Y, self._subs)
            else:
                self._oro_basis = create_grid_basis(self._model_params.ground_basis, self._X, self._Y, self._subs)
        else:
            self._oro_basis = None


class MiddleAtmosphericTemperatureMeridionalGradientDiagnostic(AtmosphericTemperatureMeridionalGradientDiagnostic):
    """Diagnostic giving the meridional gradient of the middle atmospheric temperature fields :math:`\\partial_y T_{\\rm a}` at 500hPa.
    It is identified with the meridional gradient of the baroclinic streamfunction :math:`\\partial_y \\theta_{\\rm a}` of the system.
    See also :ref:`files/model/oro_model:Mid-layer equations and the thermal wind relation` sections.

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    delta_x: float, optional
        Spatial step in the zonal direction `x` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    delta_y: float, optional
        Spatial step in the meridional direction `y` for the gridded representation of the field.
        If not provided, take an optimal guess based on the provided model's parameters.
    dimensional: bool, optional
        Indicate if the output diagnostic must be dimensionalized or not.
        Default to `True`.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, delta_x=None, delta_y=None, dimensional=True):

        AtmosphericTemperatureMeridionalGradientDiagnostic.__init__(self, model_params, delta_x, delta_y, dimensional)
        vr = self._model_params.variables_range
        self._plot_title = r'Atmospheric 500hPa Temperature Meridional Gradient'
        self._plot_units = r" (in " + self._model_params.get_variable_units(vr[0]) + r")"

        self._color_bar_format = False

    def _get_diagnostic(self, dimensional):

        natm = self._model_params.nmod[0]
        theta = np.swapaxes(self._data[natm:2 * natm, ...].T @ np.swapaxes(self._grid_basis, 0, 1), 0, 1)

        if dimensional:
            self._diagnostic_data = theta * self._model_params.temperature_scaling * 2
            self._diagnostic_data_dimensional = True
        else:
            self._diagnostic_data = theta
            self._diagnostic_data_dimensional = False
        return self._diagnostic_data


if __name__ == '__main__':
    from qgs.params.params import QgParams
    from qgs.params.params import QgParams
    from qgs.integrators.integrator import RungeKuttaIntegrator
    from qgs.functions.tendencies import create_tendencies

    pars = QgParams()
    pars.set_atmospheric_channel_fourier_modes(2, 2)
    f, Df = create_tendencies(pars)
    integrator = RungeKuttaIntegrator()
    integrator.set_func(f)
    ic = np.random.rand(pars.ndim) * 0.1
    integrator.integrate(0., 200000., 0.1, ic=ic, write_steps=5)
    time, traj = integrator.get_trajectories()
    integrator.terminate()

    theta = MiddleAtmosphericTemperatureAnomalyDiagnostic(pars)
    theta(time, traj)

    dytheta = MiddleAtmosphericTemperatureMeridionalGradientDiagnostic(pars)
    dytheta(time, traj)



"""
    Multidiagnostic class
    ======================

    This class is used to analyze and plot simultaneously several diagnostic together.

"""
import warnings
import base64
import numpy as np
from itertools import product

import matplotlib.pyplot as plt
import matplotlib.animation as animation
from IPython.display import HTML, display
from ipywidgets import interactive


class MultiDiagnostic(object):
    """class analyze and plot simultaneously several diagnostic together.
    The diagnostics information are plotted in arrays of fixed dimensions, defining the total number of diagnostics that
    the object can hold.

    Parameters
    ----------
    nrows: int
        The number of rows of diagnostic.
    ncols: int
        The number of columns of diagnostic

    Attributes
    ----------
    figure: ~matplotlib.figure.Figure
            The Matplotlib figure instance where the data are plotted.
    """

    def __init__(self, nrows, ncols):

        self._geometry = (nrows, ncols)
        self._diagnostics_list = list()
        self._diagnostics_position = list()
        self._position_to_index = np.zeros(self._geometry, dtype=int)
        self._position_occupied = np.full(self._geometry, False)
        self._diagnostic_kwargs_list = list()
        self._plot_kwargs_list = list()
        self._figures_array = None
        self._axes_array = None
        self.figure = None

    def __len__(self):
        return self._geometry[0] * self._geometry[1]

    def __call__(self, time, data):
        self.set_data(time, data)

    @property
    def _diagnostic_min_len(self):
        diag_len = list()
        for diag in self._diagnostics_list:
            diag_len.append(diag.__len__())
        try:
            min_len = min(diag_len)
            return min_len
        except:
            return 0

    @property
    def ncols(self):
        """int: The number of columns of diagnostic."""
        return self._geometry[1]

    @property
    def nrows(self):
        """int: The number of rows of diagnostic."""
        return self._geometry[0]

    @property
    def _figure_list(self):
        if self._figures_array is None:
            return None
        else:
            return list(self._figures_array.flatten())

    def set_data(self, time, data):
        """Provide the model data to all the diagnostics.

        Parameters
        ----------
        time: ~numpy.ndarray
            The time (in nondimensional timeunits) corresponding to the data.
            Its length should match the length of the last axis of the provided `data`.
        data: ~numpy.ndarray
            The model output data that the user want to convert using the diagnostic.
        """
        for diagnostic in self._diagnostics_list:
            out = diagnostic(time, data)

    @property
    def diagnostic(self):
        """list(~numpy.ndarray): The output diagnostics as a list."""
        ret = list()
        for diagnostic in self._diagnostics_list:
            diag = diagnostic.diagnostic
            ret.append(diag)
        return ret

    @property
    def diagnostics_list(self):
        """list(Diagnostic): The list of stored diagnostics."""
        return self._diagnostics_list

    @property
    def diagnostic_positions(self):
        """list(tuple(int)): Position occupied by each diagnostic in the plotting array."""
        return self._diagnostics_position

    def add_diagnostic(self, diagnostic, position=None, diagnostic_kwargs=None, plot_kwargs=None):
        """Method to add a diagnostic to the list.

        Parameters
        ----------
        diagnostic: Diagnostic
            Diagnostic to add.
        position: tuple(int), optional
            2-tuple specifying the position of the diagnostic in the plotting array. Find a *free* spot in the array if not specified.
            If the plotting array is full, a position **must** be provided to overwrite a already defined diagnostic.
        diagnostic_kwargs: dict, optional
            Dictionary of arguments to pass to the `plot`, `animate` and `movie` method of the provided diagnostic.
        plot_kwargs: dict, optional
            Specific `plot_kwargs` argument to pass to the `plot`, `animate` and `movie` method of the provided diagnostic.
            If provided, overwrite the one possibly present in the `diagnostic_kwargs` above.
        """
        if position is None:
            for i, j in product(range(self._geometry[0]), range(self._geometry[1])):
                if not self._position_occupied[i, j]:
                    position = (i, j)
                    break
            else:
                warnings.warn('All diagnostic position already occupied. Please specify the position argument to overwrite.')
                return None

            self._diagnostics_list.append(diagnostic)
            self._diagnostic_kwargs_list.append(diagnostic_kwargs)
            self._plot_kwargs_list.append(plot_kwargs)
            self._diagnostics_position.append(position)
            self._position_occupied[position[0], position[1]] = True
            self._position_to_index[position[0], position[1]] = len(self._diagnostics_list)-1
        else:
            index = self._position_to_index[position[0], position[1]]
            self._diagnostics_list[index] = diagnostic
            self._diagnostic_kwargs_list[index] = diagnostic_kwargs
            self._plot_kwargs_list[index] = plot_kwargs

    def plot(self, time_index, figure=None, figsize=(16, 9), tight_layout=False):
        """Plot the fields of the provided data at the given time index.

        Parameters
        ----------

        time_index: int
            The time index of the data.
        figure: ~matplotlib.figure.Figure, optional
            The Matplotlib figure instance where the data are plotted. Update the `figure` attribute of the object.
            If not provided, use the `figure` attribute of the object.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple. Used only if a new figure must be created.
        tight_layout: bool, optional
            Enforce a tight layout of the diagnostics axes.

        """

        if figure is None:
            self.figure = plt.figure(figsize=figsize)
            self._figures_array = self.figure.subfigures(*self._geometry, squeeze=False)
            self._axes_array = np.empty_like(self._figures_array)
            for i in range(self._geometry[0]):
                for j in range(self._geometry[1]):
                    index = self._position_to_index[i, j]
                    kwargs = self._diagnostic_kwargs_list[index]
                    if kwargs is None:
                        kwargs = dict()
                    if 'style' in kwargs:
                        if '3D' in kwargs['style']:
                            self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1, projection='3d')
                        else:
                            self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1)
                    else:
                        self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1)
        elif figure is False:
            pass
        else:
            self.figure = figure

        for i, diagnostic in enumerate(self._diagnostics_list):
            position = self._diagnostics_position[i]
            ax = self._axes_array[position[0], position[1]]
            if self._diagnostic_kwargs_list[i] is not None:
                kwargs = self._diagnostic_kwargs_list[i]
                try:
                    del kwargs['ax']
                except:
                    pass
            else:
                kwargs = dict()
            if self._plot_kwargs_list[i] is not None:
                kwargs['plot_kwargs'] = self._plot_kwargs_list[i]
            diagnostic.plot(time_index, ax=ax, **kwargs)
        if tight_layout:
            self.figure.tight_layout()

    def animate(self, output='animate', figure=None, figsize=(16, 9), stride=1, anim_kwargs=None):
        """Return the output of the `plot` method animated over time.

        Parameters
        ----------
        output: str, optional
            Define the kind of animation being created. Can be:

            * `animate`: Create and show a :class:`ipywidgets.widgets.interaction.interactive` widget. Works only in Jupyter notebooks.
            * `show`: Create and show an animation with the :mod:`matplotlib.animation` module. Works only in IPython or Python.

        figure: ~matplotlib.figure.Figure, optional
            The Matplotlib figure instance where the data are plotted. Update the `figure` attribute of the object.
            If not provided, use the `figure` attribute of the object.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple. Used only if a new figure must be created.
        stride: int, optional
            Specify the time step of the animation. Works only with `output` set to `animate`.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.
            Works only with `output` set to `show`.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or ~IPython.display.DisplayHandle
            The animation object.
        """

        if output == 'animate':
            kwargs_list = list()

            for j, diagnostic in enumerate(self._diagnostics_list):
                if self._diagnostic_kwargs_list[j] is not None:
                    kwargs_list.append(self._diagnostic_kwargs_list[j].copy())
                else:
                    kwargs_list.append(dict())
                kwargs_list[-1]['output'] = output
                kwargs_list[-1]['show'] = False
                if self._plot_kwargs_list[j] is not None:
                    kwargs_list[-1]['plot_kwargs'] = self._plot_kwargs_list[j]
                if anim_kwargs is not None:
                    kwargs_list[-1]['anim_kwargs'] = anim_kwargs

            def update_tot(time_index):
                ## no subfigure clf, so we recreate the figure all the time
                self.figure = plt.figure(figsize=figsize)
                self._figures_array = self.figure.subfigures(*self._geometry, squeeze=False)
                self._axes_array = np.empty_like(self._figures_array)
                for k in range(self._geometry[0]):
                    for j in range(self._geometry[1]):
                        index = self._position_to_index[k, j]
                        kwargs = self._diagnostic_kwargs_list[index]
                        if kwargs is None:
                            kwargs = dict()
                        if 'style' in kwargs:
                            if '3D' in kwargs['style']:
                                self._axes_array[k, j] = self._figures_array[k, j].add_subplot(1, 1, 1, projection='3d')
                            else:
                                self._axes_array[k, j] = self._figures_array[k, j].add_subplot(1, 1, 1)
                        else:
                            self._axes_array[k, j] = self._figures_array[k, j].add_subplot(1, 1, 1)

                for j, diagnostic in enumerate(self._diagnostics_list):
                    position = self._diagnostics_position[j]
                    ax = self._axes_array[position[0], position[1]]
                    kwargs_list[j]['ax'] = ax

                    update = diagnostic.animate(**kwargs_list[j])
                    update(time_index)
                plt.show()

            plot = interactive(update_tot, time_index=(0, self._diagnostic_min_len-1, stride))
            anim = display(plot)

        elif output == 'show':

            if figure is None:
                self.figure = plt.figure(figsize=figsize)
                self._figures_array = self.figure.subfigures(*self._geometry, squeeze=False)
                self._axes_array = np.empty_like(self._figures_array)
                for i in range(self._geometry[0]):
                    for j in range(self._geometry[1]):
                        index = self._position_to_index[i, j]
                        kwargs = self._diagnostic_kwargs_list[index]
                        if kwargs is None:
                            kwargs = dict()
                        if 'style' in kwargs:
                            if '3D' in kwargs['style']:
                                self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1, projection='3d')
                            else:
                                self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1)
                        else:
                            self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1)
            elif figure is False:
                pass
            else:
                self.figure = figure

            fargs_list = list()
            update_list = list()

            for j, diagnostic in enumerate(self._diagnostics_list):
                position = self._diagnostics_position[j]
                ax = self._axes_array[position[0], position[1]]
                if self._diagnostic_kwargs_list[j] is not None:
                    kwargs = self._diagnostic_kwargs_list[j].copy()
                    try:
                        del kwargs['output']
                    except:
                        pass
                    try:
                        del kwargs['show']
                    except:
                        pass
                else:
                    kwargs = dict()
                kwargs['ax'] = ax
                if self._plot_kwargs_list[j] is not None:
                    kwargs['plot_kwargs'] = self._plot_kwargs_list[j]
                if anim_kwargs is not None:
                    kwargs['anim_kwargs'] = anim_kwargs
                fig, axe, fargs, kwargs = diagnostic._init_anim(**kwargs)
                update = diagnostic._make_update(**kwargs)
                fargs_list.append(fargs)
                update_list.append(update)

            def update_tot(i, update_lst, fargs_lst):
                ax_list = list()
                for up, fags in zip(update_lst, fargs_lst):
                    out = up(i, *fags)
                    ax_list.append(out[0])
                return ax_list

            if anim_kwargs is not None:

                if 'blit' in anim_kwargs:
                    del anim_kwargs['blit']

                anim = animation.FuncAnimation(self.figure, update_tot, fargs=[update_list, fargs_list], blit=True, **anim_kwargs)

            else:
                anim = animation.FuncAnimation(self.figure, update_tot, fargs=[update_list, fargs_list], blit=True)

            plt.show()

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def movie(self, output='html', filename='', figure=None, figsize=(16, 9), anim_kwargs=None):
        """ Create and return a movie of the output of the `plot` method animated over time.

        Parameters
        ----------
        output: str, optional
            Define the kind of movie being created. Can be:

            * `jshtml`: Generate an interactive HTML representation of the animation.
            * `html5`: Generate the movie as HTML5 code.
            * `html`: Output the movie as a HTML video tag.
            * `ihtml`: Output the interactive movie as a HTML video tag.
            * `save`: Save the movie in MP4 format (H264 codec).

            Default to `html`.
        filename: str, optional
            Filename (and path) where to save the movie. Needed if `output` is set to `save`.
        figure: ~matplotlib.figure.Figure, optional
            The Matplotlib figure instance where the data are plotted. Update the `figure` attribute of the object.
            If not provided, use the `figure` attribute of the object.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple. Used only if a new figure must be created.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.
            Works only with `output` set to `show`.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or HTML code or HTML tag
            The animation object or the HTML code or tag.
        """

        if figure is None:
            self.figure = plt.figure(figsize=figsize)
            self._figures_array = self.figure.subfigures(*self._geometry, squeeze=False)
            self._axes_array = np.empty_like(self._figures_array)
            for i in range(self._geometry[0]):
                for j in range(self._geometry[1]):
                    index = self._position_to_index[i, j]
                    kwargs = self._diagnostic_kwargs_list[index]
                    if kwargs is None:
                        kwargs = dict()
                    if 'style' in kwargs:
                        if '3D' in kwargs['style']:
                            self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1, projection='3d')
                        else:
                            self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1)
                    else:
                        self._axes_array[i, j] = self._figures_array[i, j].add_subplot(1, 1, 1)
        elif figure is False:
            pass
        else:
            self.figure = figure

        fargs_list = list()
        update_list = list()

        for j, diagnostic in enumerate(self._diagnostics_list):
            position = self._diagnostics_position[j]
            ax = self._axes_array[position[0], position[1]]
            if self._diagnostic_kwargs_list[j] is not None:
                kwargs = self._diagnostic_kwargs_list[j].copy()
                try:
                    del kwargs['output']
                except:
                    pass
                try:
                    del kwargs['show']
                except:
                    pass
            else:
                kwargs = dict()
            kwargs['ax'] = ax
            if self._plot_kwargs_list[j] is not None:
                kwargs['plot_kwargs'] = self._plot_kwargs_list[j]
            if anim_kwargs is not None:
                kwargs['anim_kwargs'] = anim_kwargs
            if 'show_time' in kwargs:
                if not kwargs['show_time']:
                    kwargs['show_time_in_title'] = False
                else:
                    kwargs['show_time_in_title'] = True
                    kwargs['show_time'] = False
            else:
                kwargs['show_time_in_title'] = True
                kwargs['show_time'] = False

            fig, axe, fargs, kwargs = diagnostic._init_anim(**kwargs)
            update = diagnostic._make_update(**kwargs)
            fargs_list.append(fargs)
            update_list.append(update)

        def update_tot(i, update_lst, fargs_lst):
            ax_list = list()
            for up, fags in zip(update_lst, fargs_lst):
                out = up(i, *fags)
                ax_list.append(out[0])
            return ax_list

        if anim_kwargs is not None:

            if 'blit' in anim_kwargs:
                del anim_kwargs['blit']

            anim = animation.FuncAnimation(self.figure, update_tot, fargs=[update_list, fargs_list], blit=False, **anim_kwargs)

        else:
            anim = animation.FuncAnimation(self.figure, update_tot, fargs=[update_list, fargs_list], blit=False)

        if 'html' in output:

            if output == "jshtml" or output == 'ihtml':
                jshtml = anim.to_jshtml()
                if output == "jshtml":
                    return jshtml
                else:
                    return HTML(jshtml)
            else:
                html5 = anim.to_html5_video()
                if output == 'html5':
                    return html5
                else:
                    return HTML(html5)

        elif output == 'save':

            if not filename:
                warnings.warn('No filename provided to the method animate. Video not saved !\n Please provide a filename.')

            html = anim.to_html5_video()
            start_index = html.index('base64,')
            start_index += len('base64,')
            end_index = html.index('">', start_index)
            video = html[start_index: end_index]
            with open(filename, 'wb') as f:
                f.write(base64.b64decode(video))

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim


class FieldsDiagnosticsList(object):
    """General base class for plotting multiple diagnostics on a single axe. The diagnostics must be provided as a list.
    Assumes that the first diagnostic in the list set the parameters that are not specified.

    Parameters
    ----------
    diagnostics_list: list
        List of initialized auxialiary diagnostics to plot with the main diagnostic.
    """

    def __init__(self, diagnostics_list=None):

        if diagnostics_list is not None:
            self._diagnostics_list = diagnostics_list
        else:
            self._diagnostics_list = list()

    def append_diagnostic(self, diagnostic):
        """Method to add an auxiliary diagnostic to the list.

        Parameters
        ----------
        diagnostic: Diagnostic
            The diagnostic to add to the list.
        """
        self._diagnostics_list.append(diagnostic)

    def __call__(self, time, data, index=None):
        self.set_data(time, data, index)

    def __len__(self):
        diag_len = list()
        for diag in self._diagnostics_list:
            diag_len.append(diag.__len__())
        try:
            min_len = min(diag_len)
            return min_len
        except:
            return 0

    def set_data(self, time, data, index=None):
        """Provide the model data to the index-th diagnostic.

        Parameters
        ----------
        time: ~numpy.ndarray
            The time (in nondimensional timeunits) corresponding to the data.
            Its length should match the length of the last axis of the provided `data`.
        data: ~numpy.ndarray
            The model output data that the user want to convert using the diagnostic.
            Should be a 2D array of shape (:attr:`~.params.QgParams.ndim`, number_of_timesteps).
        index: int or None
            The index of the diagnostic in the list to provide the data to.
        """
        if self._diagnostics_list is None:
            warnings.warn('No diagnostics available. Doing nothing.')
            return None

        if index is None:
            return None
        else:
            self._diagnostics_list[index].set_data(time, data)

    def plot(self, time_index, style="image", ax=None, figsize=(16, 9),
             contour_labels=True, color_bar=True, show_time=True, plot_kwargs=None, oro_kwargs=None):
        """Plot the field of the provided diagnostics at the given time index.
        Almost all the parameters (except `ax` and `figsize`) and fig should be lists corresponding to diagnostics in the list.
        If a single parameter is provided, it applies to all the diagnostics.

        Parameters
        ----------
        time_index: list(int)
            The time index of the data.
        style: list(str), optional
            The style of the plot. Can be:

            * `image`: show the fields as images with a given colormap specified in the `plot_kwargs` argument.
            * `contour`: show the fields as contour superimposed on the image of the orographic height (if it exists, see the `oro_kwargs` below).
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        contour_labels: list(bool), optional
            If `style` is set to `contour`, specify if the contours must be labelled with their value or not.
            Default to `True`.
        color_bar: list(bool), optional
            Specify if a color bar must be drawn beside the plot or not.
            Default to `True`.
        show_time: list(bool), optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        plot_kwargs: list(dict), optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.
        oro_kwargs: list(dict), optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method plotting the image of the orography if `style` is set to `contour`.

        Returns
        -------
        ~matplotlib.axes.Axes
            An axes where the data were plotted.
        """

        if self._diagnostics_list is None:
            warnings.warn('No diagnostics available. Showing nothing.')
            return None

        if not isinstance(time_index, (list, tuple)):
            time_index = (len(self._diagnostics_list)) * [time_index]

        if not isinstance(style, (list, tuple)):
            style = (len(self._diagnostics_list)) * [style]

        if not isinstance(contour_labels, (list, tuple)):
            contour_labels = (len(self._diagnostics_list)) * [contour_labels]

        if not isinstance(color_bar, (list, tuple)):
            color_bar = (len(self._diagnostics_list)) * [color_bar]

        if not isinstance(show_time, (list, tuple)):
            show_time = (len(self._diagnostics_list)) * [show_time]

        if not isinstance(plot_kwargs, (list, tuple)):
            plot_kwargs = (len(self._diagnostics_list)) * [plot_kwargs]

        if not isinstance(oro_kwargs, (list, tuple)):
            oro_kwargs = (len(self._diagnostics_list)) * [oro_kwargs]

        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.gca()

        for j, diagnostic in enumerate(self._diagnostics_list):
            diagnostic.plot(time_index[j], style[j], ax, figsize, contour_labels[j], color_bar[j], show_time[j], plot_kwargs[j], oro_kwargs[j])

        return ax

    def movie(self, output='html', filename='', style="image", ax=None, figsize=(16, 9),
              contour_labels=True, color_bar=True, show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None):
        """ Create and return a movie of the output of the `plot` method animated over time.
        Almost all the parameters (except `output`, `filename`, `ax`, `figsize`, and `anim_kwargs`) and fig should be lists corresponding to diagnostics in the list.
        If a single parameter is provided, it applies to all the diagnostics.

        Parameters
        ----------
        output: str, optional
            Define the kind of movie being created. Can be:

            * `jshtml`: Generate an interactive HTML representation of the animation.
            * `html5`: Generate the movie as HTML5 code.
            * `html`: Output the movie as a HTML video tag.
            * `ihtml`: Output the interactive movie as a HTML video tag.
            * `save`: Save the movie in MP4 format (H264 codec).

            Default to `html`.
        filename: str, optional
            Filename (and path) where to save the movie. Needed if `output` is set to `save`.
        style: lits(str), optional
            The style of the plot. Can be:

            * `image`: show the fields as images with a given colormap specified in the `plot_kwargs` argument.
            * `contour`: show the fields as contour superimposed on the image of the orographic height (see the `oro_kwargs` below).
        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        contour_labels: list(bool), optional
            If `style` is set to `contour`, specify if the contours must be labelled with their value or not.
            Default to `True`.
        color_bar: list(bool), optional
            Specify if a color bar must be drawn beside the plot or not.
            Default to `True`.
        show_time: list(bool), optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        plot_kwargs: list(dict), optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.
        oro_kwargs: list(dict), optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method plotting the image of the orography if `style` is set to `contour`.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or HTML code or HTML tag
            The animation object or the HTML code or tag.
        """

        anim = self._make_anim(style, ax, figsize, contour_labels, color_bar, show_time, plot_kwargs, oro_kwargs, anim_kwargs, False)

        if 'html' in output:

            if output == "jshtml" or output == 'ihtml':
                jshtml = anim.to_jshtml()
                if output == "jshtml":
                    return jshtml
                else:
                    return HTML(jshtml)
            else:
                html5 = anim.to_html5_video()
                if output == 'html5':
                    return html5
                else:
                    return HTML(html5)

        elif output == 'save':

            if not filename:
                warnings.warn('No filename provided to the method animate. Video not saved !\n Please provide a filename.')

            html5 = anim.to_html5_video()
            start_index = html5.index('base64,')
            start_index += len('base64,')
            end_index = html5.index('">', start_index)
            video = html5[start_index: end_index]
            with open(filename, 'wb') as f:
                f.write(base64.b64decode(video))
            return html5

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def animate(self, output='animate',  style="image", ax=None, figsize=(16, 9),
                contour_labels=True, color_bar=True, show_time=True, stride=1, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None, show=True):
        """Return the output of the `plot` method animated over time.
        Almost all the parameters (except `animate`, `ax`, `figsize`, `stride`, `anim_kwargs` and `show`) and fig should be lists corresponding to diagnostics in the list.
        If a single parameter is provided, it applies to all the diagnostics.

        Parameters
        ----------
        output: str, optional
            Define the kind of animation being created. Can be:

            * `animate`: Create and show a :class:`ipywidgets.widgets.interaction.interactive` widget. Works only in Jupyter notebooks.
            * `show`: Create and show an animation with the :mod:`matplotlib.animation` module. Works only in IPython or Python.

        style: list(str), optional
            The style of the plot. Can be:

            * `image`: show the fields as images with a given colormap specified in the `plot_kwargs` argument.
            * `contour`: show the fields as contour superimposed on the image of the orographic height (see the `oro_kwargs` below).

        ax: ~matplotlib.axes.Axes, optional
            An axes on which to plot the fields.
        figsize: tuple(float), optional
            The size of the figure in inches as a 2-tuple.
        contour_labels: list(bool), optional
            If `style` is set to `contour`, specify if the contours must be labelled with their value or not.
            Default to `True`.
        color_bar: list(bool), optional
            Specify if a color bar must be drawn beside the plot or not.
            Default to `True`.
        show_time: list(bool), optional
            Show the timestamp of the field on the plot or not.
            Default to `True`.
        stride: int, optional
            Specify the time step of the animation. Works only with `output` set to `animate`.
        plot_kwargs: list(dict), optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method if `style` is set to `image`, or to the :meth:`matplotlib.axes.Axes.contour` method if `style` is set to `contour`.
        oro_kwargs: list(dict), optional
            Arguments to pass to the :meth:`matplotlib.axes.Axes.imshow` method plotting the image of the orography if `style` is set to `contour`.
        anim_kwargs: dict, optional
            Arguments to pass to the :class:`matplotlib.animation.FuncAnimation` instantiation method. Specify the parameters of the animation.
            Works only with `output` set to `show`.
        show: bool, optional
            Whether to plot or not the animation.

        Returns
        -------
        ~matplotlib.animation.FuncAnimation or ~IPython.display.DisplayHandle or callable
            The animation object or the callable to update the widget, depending on the value of the `output` and `show` parameters.
        """

        if self._diagnostics_list is None:
            warnings.warn('No diagnostic available. Showing nothing. Returning None.')
            return None

        if not isinstance(style, (list, tuple)):
            style = (len(self._diagnostics_list)) * [style]

        if not isinstance(contour_labels, (list, tuple)):
            contour_labels = (len(self._diagnostics_list)) * [contour_labels]

        if not isinstance(color_bar, (list, tuple)):
            color_bar = (len(self._diagnostics_list)) * [color_bar]

        if not isinstance(show_time, (list, tuple)):
            show_time = (len(self._diagnostics_list)) * [show_time]

        if not isinstance(plot_kwargs, (list, tuple)):
            plot_kwargs = (len(self._diagnostics_list)) * [plot_kwargs]

        if not isinstance(oro_kwargs, (list, tuple)):
            oro_kwargs = (len(self._diagnostics_list)) * [oro_kwargs]

        if output == 'animate':

            if style == 'image':

                for i, diagnostic in enumerate(self._diagnostics_list):

                    vmin = diagnostic.min() * 1.03
                    vmax = diagnostic.max() * 1.03
                    if plot_kwargs[i] is not None:
                        if 'vmin' not in plot_kwargs[i]:
                            plot_kwargs[i]['vmin'] = vmin
                        if 'vmax' not in plot_kwargs[i]:
                            plot_kwargs[i]['vmax'] = vmax
                    else:
                        plot_kwargs[i] = dict()
                        plot_kwargs[i]['vmin'] = vmin
                        plot_kwargs[i]['vmax'] = vmax

            def update(time_index):
                axe = None
                for i, diagnostic in enumerate(self._diagnostics_list):
                    if i == 0:
                        pack = diagnostic.plot(time_index, style[i], ax, figsize, contour_labels[i], color_bar[i], show_time[i], plot_kwargs[i], oro_kwargs[i])
                        if hasattr(pack, '__getitem__'):
                            axe = pack[0]
                        else:
                            axe = pack
                    else:
                        diagnostic.plot(time_index, style[i], axe, figsize, contour_labels[i], color_bar[i], show_time[i], plot_kwargs[i], oro_kwargs[i])
                if show:
                    plt.show()

            if show:
                plot = interactive(update, time_index=(0, len(self._diagnostics_list[0])-1, stride))
                anim = display(plot)
            else:
                return update

        elif output == 'show':

            anim = self._make_anim(style, ax, figsize, contour_labels, color_bar, show_time, plot_kwargs, oro_kwargs, anim_kwargs, True)
            if show:
                plt.show()

        else:
            warnings.warn('Provided output parameter ' + output + ' not supported ! Nothing to plot. Returning None.')
            anim = None

        return anim

    def _init_anim(self, style="image", ax=None, figsize=(16, 9), contour_labels=True, color_bar=True,
                   show_time_in_title=False, show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None):

        if self._diagnostics_list is None:
            warnings.warn('No diagnostic data available. Showing nothing. Returning None.')
            return None

        if not isinstance(style, (list, tuple)):
            style = (len(self._diagnostics_list)) * [style]

        if not isinstance(contour_labels, (list, tuple)):
            contour_labels = (len(self._diagnostics_list)) * [contour_labels]

        if not isinstance(color_bar, (list, tuple)):
            color_bar = (len(self._diagnostics_list)) * [color_bar]

        if not isinstance(show_time, (list, tuple)):
            show_time = (len(self._diagnostics_list)) * [show_time]

        if not isinstance(plot_kwargs, (list, tuple)):
            plot_kwargs = (len(self._diagnostics_list)) * [plot_kwargs]

        if not isinstance(oro_kwargs, (list, tuple)):
            oro_kwargs = (len(self._diagnostics_list)) * [oro_kwargs]

        if style == 'image':

            for i, diagnostic in enumerate(self._diagnostics_list):

                vmin = diagnostic.min() * 1.03
                vmax = diagnostic.max() * 1.03
                if plot_kwargs[i] is not None:
                    if 'vmin' not in plot_kwargs[i]:
                        plot_kwargs[i]['vmin'] = vmin
                    if 'vmax' not in plot_kwargs[i]:
                        plot_kwargs[i]['vmax'] = vmax
                else:
                    plot_kwargs[i] = dict()
                    plot_kwargs[i]['vmin'] = vmin
                    plot_kwargs[i]['vmax'] = vmax

        for i, diagnostic in enumerate(self._diagnostics_list):
            pack = diagnostic.plot(0, style[i], ax, figsize, contour_labels[i], color_bar[i], show_time_in_title, plot_kwargs[i], oro_kwargs[i])
            if hasattr(pack, '__getitem__'):
                ax = pack[0]
            else:
                ax = pack
            fig = ax.figure

            if show_time[i]:
                if self._diagnostics_list[i].dimensional:
                    tt = " at " + "{:.2f}".format(self._diagnostics_list[i]._model_params.dimensional_time * self._diagnostics_list[i]._time[0]) \
                         + " " + self._diagnostics_list[i]._model_params.time_unit
                else:
                    tt = " at " + str(self._diagnostics_list[i]._time[0]) + " timeunits"
                ax.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

        fargs = (ax, show_time_in_title, show_time)

        kwargs = {'style': style, 'ax': ax, 'figsize': figsize, 'contour_labels': contour_labels, 'color_bar': color_bar,
                  'show_time_in_title': show_time_in_title, 'show_time': show_time, 'plot_kwargs': plot_kwargs, 'oro_kwargs': oro_kwargs,
                  'anim_kwargs': anim_kwargs}

        return fig, ax, fargs, kwargs

    def _make_update(self, style="image", ax=None, figsize=(16, 9), contour_labels=True, color_bar=True,
                     show_time_in_title=False, show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None):
        def update(i, axe, show_t_in_title, show_t):
            axe.clear()
            for j, diagnostic in enumerate(self._diagnostics_list):
                axe = diagnostic.plot(i, style[j], axe, figsize, contour_labels[j], False, show_t_in_title, plot_kwargs[j], oro_kwargs[j])
                if show_t[j]:
                    if self._diagnostics_list[j].dimensional:
                        tt = " at " + "{:.2f}".format(self._diagnostics_list[j]._model_params.dimensional_time * self._diagnostics_list[j]._time[i]) \
                             + " " + self._diagnostics_list[j]._model_params.time_unit
                    else:
                        tt = " at " + str(self._diagnostics_list[j]._time[i]) + " timeunits"
                    axe.text(0.1, 0.9, tt, horizontalalignment='center', verticalalignment='center', transform=axe.transAxes)
            return [axe]

        return update

    def _make_anim(self, style="image", ax=None, figsize=(16, 9), contour_labels=True, color_bar=True,
                   show_time=True, plot_kwargs=None, oro_kwargs=None, anim_kwargs=None, blit=True):

        if show_time:
            if blit:
                show_time_in_title = False
            else:
                show_time_in_title = True
                show_time = False
        else:
            show_time_in_title = False

        fig, ax, fargs, kwargs = self._init_anim(style, ax, figsize, contour_labels, color_bar, show_time_in_title, show_time, plot_kwargs, oro_kwargs, anim_kwargs)

        update = self._make_update(**kwargs)

        if anim_kwargs is not None:

            if 'blit' in anim_kwargs:
                del anim_kwargs['blit']

            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit, **anim_kwargs)

        else:
            anim = animation.FuncAnimation(fig, update, fargs=fargs, blit=blit)

        return anim


if __name__ == '__main__':
    from qgs.params.params import QgParams
    from qgs.params.params import QgParams
    from qgs.integrators.integrator import RungeKuttaIntegrator
    from qgs.functions.tendencies import create_tendencies
    from qgs.diagnostics.streamfunctions import LowerLayerAtmosphericStreamfunctionDiagnostic, UpperLayerAtmosphericStreamfunctionDiagnostic, \
        MiddleAtmosphericStreamfunctionDiagnostic
    from qgs.diagnostics.temperatures import MiddleAtmosphericTemperatureDiagnostic
    from qgs.diagnostics.variables import VariablesDiagnostic, GeopotentialHeightDifferenceDiagnostic

    pars = QgParams()
    pars.set_atmospheric_channel_fourier_modes(2, 2)
    f, Df = create_tendencies(pars)
    integrator = RungeKuttaIntegrator()
    integrator.set_func(f)
    ic = np.random.rand(pars.ndim) * 0.1
    integrator.integrate(0., 200000., 0.1, ic=ic, write_steps=5)
    time, traj = integrator.get_trajectories()
    integrator.terminate()

    psi3 = LowerLayerAtmosphericStreamfunctionDiagnostic(pars)
    psi1 = UpperLayerAtmosphericStreamfunctionDiagnostic(pars)
    psi = MiddleAtmosphericStreamfunctionDiagnostic(pars)
    theta = MiddleAtmosphericTemperatureDiagnostic(pars)

    var_nondim = VariablesDiagnostic([1, 2, 0], pars, False)
    geo_dim = GeopotentialHeightDifferenceDiagnostic([[[np.pi/pars.scale_params.n, np.pi/4], [np.pi/pars.scale_params.n, 3*np.pi/4]],
                                                      [[0, np.pi/4], [0, 3*np.pi/4]]],
                                                     pars, True)

    kw = {'frames': 100, 'interval': 100, 'blit': False}
    pkw = {'ms': 0.1}

    m = MultiDiagnostic(2, 2)
    m.add_diagnostic(var_nondim, diagnostic_kwargs={'style':'2Dscatter'}, plot_kwargs=pkw)
    # m.add_diagnostic(geo_dim, diagnostic_kwargs={'style': 'moving-timeserie'}, plot_kwargs=pkw)
    m.add_diagnostic(psi1)
    m.add_diagnostic(psi3)
    m.add_diagnostic(theta)
    m(time, traj)



"""
    Differential diagnostic base class
    ==================================

    Abstract base classes defining diagnostics on differentiated grids.

    Description of the classes
    --------------------------

    * :class:`DifferentialFieldDiagnostic`: Base class for diagnostics returning model's fields based on differential grids.
    * :class:`LaplacianFieldDiagnostic`: Base class for diagnostics returning model's fields based on Laplacian grids.

    Warnings
    --------

    These are `abstract base class`_, they must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

"""

from qgs.diagnostics.util import create_grid_basis

from qgs.diagnostics.base import FieldDiagnostic


class DifferentialFieldDiagnostic(FieldDiagnostic):
    """General base class for differential fields diagnostic.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, dimensional):

        FieldDiagnostic.__init__(self, model_params, dimensional)

    def _configure_differential_grid(self, basis, derivative, order, delta_x=None, delta_y=None):

        self._compute_grid(delta_x, delta_y)

        if derivative == "dx":
            dx_basis = basis.x_derivative(order)
            self._grid_basis = create_grid_basis(dx_basis, self._X, self._Y, self._subs)

        elif derivative == "dy":
            dy_basis = basis.y_derivative(order)
            self._grid_basis = create_grid_basis(dy_basis, self._X, self._Y, self._subs)


class LaplacianFieldDiagnostic(FieldDiagnostic):
    """General base class for Laplacian fields diagnostic.
    This is an `abstract base class`_, it must be subclassed to create new diagnostics!

    .. _abstract base class: https://docs.python.org/3/glossary.html#term-abstract-base-class

    Parameters
    ----------

    model_params: QgParams
        An instance of the model parameters.
    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.

    Attributes
    ----------

    dimensional: bool
        Indicate if the output diagnostic must be dimensionalized or not.
    """

    def __init__(self, model_params, dimensional):

        FieldDiagnostic.__init__(self, model_params, dimensional)

    def _configure_laplacian_grid(self, basis, delta_x=None, delta_y=None):

        self._compute_grid(delta_x, delta_y)

        dx_basis = basis.x_derivative(2)
        dx_grid = create_grid_basis(dx_basis, self._X, self._Y, self._subs)

        dy_basis = basis.y_derivative(2)
        dy_grid = create_grid_basis(dy_basis, self._X, self._Y, self._subs)

        self._grid_basis = dx_grid + dy_grid
